@@ -15,6 +15,8 @@
 #include "nls.h"
 #include "blkdev.h"
 #include "pathnames.h"
+#include "closestream.h"
+#include "sysfs.h"
 
 struct bdc {
 	long		ioc;		/* ioctl code */
@@ -126,7 +128,7 @@ static const struct bdc bdcms[] =
 		.argname = "<bytes>",
 		.argtype = ARG_INT,
 		.flags = FL_NORESULT,
-	        .help = N_("set blocksize")
+	        .help = N_("set blocksize on file descriptor opening the block device")
 	},{
 		IOCTL_ENTRY(BLKGETSIZE),
 		.name = "--getsize",
@@ -176,27 +178,42 @@ static const struct bdc bdcms[] =
 	}
 };
 
-static void __attribute__ ((__noreturn__)) usage(FILE * out)
+static void __attribute__((__noreturn__)) usage(void)
 {
 	size_t i;
-	fprintf(out, _("\nUsage:\n"
-		       " %1$s -V\n"
-		       " %1$s --report [devices]\n"
-		       " %1$s [-v|-q] commands devices\n\n"
-		       "Available commands:\n"), program_invocation_short_name);
 
-	fprintf(out, _(" %-25s get size in 512-byte sectors\n"), "--getsz");
+	fputs(USAGE_HEADER, stdout);
+	printf(_(
+	         " %1$s [-v|-q] commands devices\n"
+	         " %1$s --report [devices]\n"
+	         " %1$s -h|-V\n"
+		), program_invocation_short_name);
+
+	fputs(USAGE_SEPARATOR, stdout);
+	puts(  _("Call block device ioctls from the command line."));
+
+	fputs(USAGE_OPTIONS, stdout);
+	puts(  _(" -q             quiet mode"));
+	puts(  _(" -v             verbose mode"));
+	puts(  _("     --report   print report for specified (or all) devices"));
+	fputs(USAGE_SEPARATOR, stdout);
+	printf(USAGE_HELP_OPTIONS(16));
+
+	fputs(USAGE_SEPARATOR, stdout);
+	puts(  _("Available commands:"));
+	printf(_(" %-25s get size in 512-byte sectors\n"), "--getsz");
 	for (i = 0; i < ARRAY_SIZE(bdcms); i++) {
 		if (bdcms[i].argname)
-			fprintf(out, " %s %-*s %s\n", bdcms[i].name,
+			printf(" %s %-*s %s\n", bdcms[i].name,
 				(int)(24 - strlen(bdcms[i].name)),
 				bdcms[i].argname, _(bdcms[i].help));
 		else
-			fprintf(out, " %-25s %s\n", bdcms[i].name,
+			printf(" %-25s %s\n", bdcms[i].name,
 				_(bdcms[i].help));
 	}
-	fputc('\n', out);
-	exit(out == stderr ? EXIT_FAILURE : EXIT_SUCCESS);
+
+	printf(USAGE_MAN_TAIL("blockdev(8)"));
+	exit(EXIT_SUCCESS);
 }
 
 static int find_cmd(char *s)
@@ -209,10 +226,10 @@ static int find_cmd(char *s)
 	return -1;
 }
 
-void do_commands(int fd, char **argv, int d);
-void report_header(void);
-void report_device(char *device, int quiet);
-void report_all_devices(void);
+static void do_commands(int fd, char **argv, int d);
+static void report_header(void);
+static void report_device(char *device, int quiet);
+static void report_all_devices(void);
 
 int main(int argc, char **argv)
 {
@@ -221,18 +238,20 @@ int main(int argc, char **argv)
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
+	atexit(close_stdout);
 
-	if (argc < 2)
-		usage(stderr);
+	if (argc < 2) {
+		warnx(_("not enough arguments"));
+		errtryhelp(EXIT_FAILURE);
+	}
 
 	/* -V not together with commands */
 	if (!strcmp(argv[1], "-V") || !strcmp(argv[1], "--version")) {
-		printf(_("%s (%s)\n"), program_invocation_short_name,
-		       PACKAGE_STRING);
+		printf(UTIL_LINUX_VERSION);
 		return EXIT_SUCCESS;
 	}
 	if (!strcmp(argv[1], "-h") || !strcmp(argv[1], "--help"))
-		usage(stdout);
+		usage();
 
 	/* --report not together with other commands */
 	if (!strcmp(argv[1], "--report")) {
@@ -265,8 +284,10 @@ int main(int argc, char **argv)
 			break;
 	}
 
-	if (d >= argc)
-		usage(stderr);
+	if (d >= argc) {
+		warnx(_("no device specified"));
+		errtryhelp(EXIT_FAILURE);
+	}
 
 	for (k = d; k < argc; k++) {
 		fd = open(argv[k], O_RDONLY, 0);
@@ -278,16 +299,16 @@ int main(int argc, char **argv)
 	return EXIT_SUCCESS;
 }
 
-void do_commands(int fd, char **argv, int d)
+static void do_commands(int fd, char **argv, int d)
 {
 	int res, i, j;
-	int iarg;
-	unsigned int uarg;
-	unsigned short huarg;
-	long larg;
-	long long llarg;
-	unsigned long lu;
-	unsigned long long llu;
+	int iarg = 0;
+	unsigned int uarg = 0;
+	unsigned short huarg = 0;
+	long larg = 0;
+	long long llarg = 0;
+	unsigned long lu = 0;
+	unsigned long long llu = 0;
 	int verbose = 0;
 
 	for (i = 1; i < d; i++) {
@@ -313,7 +334,7 @@ void do_commands(int fd, char **argv, int d)
 		j = find_cmd(argv[i]);
 		if (j == -1) {
 			warnx(_("Unknown command: %s"), argv[i]);
-			usage(stderr);
+			errtryhelp(EXIT_FAILURE);
 		}
 
 		switch (bdcms[j].argtype) {
@@ -330,7 +351,7 @@ void do_commands(int fd, char **argv, int d)
 				if (i == d - 1) {
 					warnx(_("%s requires an argument"),
 					      bdcms[j].name);
-					usage(stderr);
+					errtryhelp(EXIT_FAILURE);
 				}
 				iarg = atoi(argv[++i]);
 			} else
@@ -363,7 +384,7 @@ void do_commands(int fd, char **argv, int d)
 		}
 
 		if (res == -1) {
-			perror(bdcms[j].iocname);
+			warn(_("ioctl error on %s"), bdcms[j].iocname);
 			if (verbose)
 				printf(_("%s failed.\n"), _(bdcms[j].help));
 			exit(EXIT_FAILURE);
@@ -405,11 +426,11 @@ void do_commands(int fd, char **argv, int d)
 	}
 }
 
-void report_all_devices(void)
+static void report_all_devices(void)
 {
 	FILE *procpt;
 	char line[200];
-	char ptname[200];
+	char ptname[200 + 1];
 	char device[210];
 	int ma, mi, sz;
 
@@ -429,13 +450,14 @@ void report_all_devices(void)
 	fclose(procpt);
 }
 
-void report_device(char *device, int quiet)
+static void report_device(char *device, int quiet)
 {
 	int fd;
 	int ro, ssz, bsz;
 	long ra;
 	unsigned long long bytes;
-	struct hd_geometry g;
+	uint64_t start = 0;
+	struct stat st;
 
 	fd = open(device, O_RDONLY | O_NONBLOCK);
 	if (fd < 0) {
@@ -445,15 +467,27 @@ void report_device(char *device, int quiet)
 	}
 
 	ro = ssz = bsz = 0;
-	g.start = ra = 0;
+	ra = 0;
+	if (fstat(fd, &st) == 0 && !sysfs_devno_is_wholedisk(st.st_rdev)) {
+		struct sysfs_cxt cxt;
+
+		if (sysfs_init(&cxt, st.st_rdev, NULL))
+			err(EXIT_FAILURE,
+				_("%s: failed to initialize sysfs handler"),
+				device);
+		if (sysfs_read_u64(&cxt, "start", &start))
+			err(EXIT_FAILURE,
+				_("%s: failed to read partition start from sysfs"),
+				device);
+		sysfs_deinit(&cxt);
+	}
 	if (ioctl(fd, BLKROGET, &ro) == 0 &&
 	    ioctl(fd, BLKRAGET, &ra) == 0 &&
 	    ioctl(fd, BLKSSZGET, &ssz) == 0 &&
 	    ioctl(fd, BLKBSZGET, &bsz) == 0 &&
-	    ioctl(fd, HDIO_GETGEO, &g) == 0 &&
 	    blkdev_get_size(fd, &bytes) == 0) {
-		printf("%s %5ld %5d %5d %10ld %15lld   %s\n",
-		       ro ? "ro" : "rw", ra, ssz, bsz, g.start, bytes, device);
+		printf("%s %5ld %5d %5d %10ju %15lld   %s\n",
+		       ro ? "ro" : "rw", ra, ssz, bsz, start, bytes, device);
 	} else {
 		if (!quiet)
 			warnx(_("ioctl error on %s"), device);
@@ -462,7 +496,7 @@ void report_device(char *device, int quiet)
 	close(fd);
 }
 
-void report_header()
+static void report_header(void)
 {
 	printf(_("RO    RA   SSZ   BSZ   StartSec            Size   Device\n"));
 }
